import numpy as np


class Units:
    def __init__(self):
        self._BASE_LENGTH_UNIT = "m"
        self._BASE_AREA_UNIT = "m2"
        self._BASE_ANGLE_UNIT = "rad"

        self._LENGTH_UNIT = self._BASE_LENGTH_UNIT
        self._AREA_UNIT = self._BASE_AREA_UNIT
        self._ANGLE_UNIT = self._BASE_ANGLE_UNIT

        self._LENGTH_CONVERSIONS = {"in": 39.37007874015748, "mm": 1000., "cm": 100.}
        self._AREA_CONVERSIONS = {"in2": 1550.0031000062002, "mm2": 1.0e6, "cm2": 1.0e4}
        self._ANGLE_CONVERSIONS = {"deg": 57.29577951308232}

    def current_length_unit(self):
        return self._LENGTH_UNIT

    def set_current_length_unit(self, unit: str):
        self._LENGTH_UNIT = unit

    def convert_length_to_base(self, value: float, unit: str):
        if unit == self._BASE_LENGTH_UNIT:
            return value
        return value / self._LENGTH_CONVERSIONS[unit]

    def convert_length_from_base(self, value: float, unit: str):
        if unit == self._BASE_LENGTH_UNIT:
            return value
        return value * self._LENGTH_CONVERSIONS[unit]

    def current_area_unit(self):
        return self._AREA_UNIT

    def set_current_area_unit(self, unit: str):
        self._AREA_UNIT = unit

    def convert_area_to_base(self, value: float, unit: str):
        if unit == self._BASE_AREA_UNIT:
            return value
        return value / self._AREA_CONVERSIONS[unit]

    def convert_area_from_base(self, value: float, unit: str):
        if unit == self._BASE_AREA_UNIT:
            return value
        return value * self._AREA_CONVERSIONS[unit]

    def current_angle_unit(self):
        return self._ANGLE_UNIT

    def set_current_angle_unit(self, unit: str):
        self._ANGLE_UNIT = unit

    def convert_angle_to_base(self, value: float, unit: str):
        if unit == self._BASE_ANGLE_UNIT:
            return value
        return value / self._ANGLE_CONVERSIONS[unit]

    def convert_angle_from_base(self, value: float, unit: str):
        if unit == self._BASE_ANGLE_UNIT:
            return value
        return value * self._ANGLE_CONVERSIONS[unit]

    def pi(self):
        return self.convert_angle_from_base(np.pi, self.current_angle_unit())

    def sin(self, angle: float):
        return np.sin(self.convert_angle_to_base(angle, self.current_angle_unit()))

    def cos(self, angle: float):
        return np.cos(self.convert_angle_to_base(angle, self.current_angle_unit()))



import typing
from dataclasses import dataclass
from abc import abstractmethod

import numpy as np

from pymead.core.constraint_equations import measure_rel_angle3, measure_point_line_distance_unsigned
from pymead.core.line import PolyLine
from pymead.core.param import Param, AngleParam, LengthParam
from pymead.core.point import Point
from pymead.core.pymead_obj import PymeadObj


class GeoCon(PymeadObj):

    default_name: str = ""

    def __init__(self, param: Param or None, child_nodes: list, kind: str,
                 name: str or None = None, secondary_params: typing.List[Param] = None):
        self._param = None
        self.set_param(param)
        sub_container = "geocon"
        super().__init__(sub_container=sub_container)
        name = self.default_name if name is None else name
        self.set_name(name)
        if self.param() is not None:
            self.param().geo_cons.append(self)
            if self.param().name() == "unnamed":
                self.param().set_name(f"{self.name()}.par")
        self.child_nodes = child_nodes
        self.kind = kind
        self.secondary_params = [] if secondary_params is None else secondary_params
        self.data = None
        self.add_constraint_to_points()

    def param(self):
        return self._param

    def set_param(self, param: Param):
        self._param = param

    def add_constraint_to_points(self):
        for child_node in self.child_nodes:
            if not isinstance(child_node, Point) or self in child_node.geo_cons:
                return

            child_node.geo_cons.append(self)

    def remove_constraint_from_points(self):
        for child_node in self.child_nodes:
            if not isinstance(child_node, Point) or self not in child_node.geo_cons:
                return

            child_node.geo_cons.remove(self)

    @abstractmethod
    def verify(self) -> bool:
        pass


class DistanceConstraint(GeoCon):

    default_name = "DistCon-1"

    def __init__(self, p1: Point, p2: Point, value: float or LengthParam = None, name: str = None):
        self.p1 = p1
        self.p2 = p2
        value = self.p1.measure_distance(self.p2) if value is None else value
        param = value if isinstance(value, Param) else LengthParam(value=value, name="Length-1")
        super().__init__(param=param, name=name, child_nodes=[self.p1, self.p2], kind="d")

    def __repr__(self):
        return f"{self.__class__.__name__} {self.name()}<v={self.param().value()}>"

    def get_dict_rep(self) -> dict:
        return {"p1": self.p1.name(), "p2": self.p2.name(), "value": self.param().name(),
                "constraint_type": self.__class__.__name__}

    def verify(self) -> bool:
        measured_distance = self.p1.measure_distance(self.p2)
        return np.isclose(measured_distance, self.param().value(), rtol=1e-14)


class AntiParallel3Constraint(GeoCon):

    default_name = "AntiPar3Con-1"

    def __init__(self, p1: Point, p2: Point, p3: Point, name: str = None, polyline: PolyLine = None,
                 point_on_curve: Point = None):
        self.p1 = p1
        self.polyline = polyline
        self.point_on_curve = point_on_curve
        if polyline and polyline not in p1.curves and point_on_curve is p1:
            p1.curves.append(polyline)
        self.p2 = p2
        self.p3 = p3
        if polyline and polyline not in p3.curves and point_on_curve is p3:
            p3.curves.append(polyline)
        super().__init__(param=None, name=name, child_nodes=[self.p1, self.p2, self.p3], kind="a3")

    def __repr__(self):
        return f"{self.__class__.__name__} {self.name()}"

    def get_dict_rep(self) -> dict:
        return {"p1": self.p1.name(), "p2": self.p2.name(), "p3": self.p3.name(),
                "constraint_type": self.__class__.__name__,
                "polyline": self.polyline.name() if self.polyline is not None else None,
                "point_on_curve": self.point_on_curve.name() if self.point_on_curve is not None else None}

    def verify(self) -> bool:
        a1 = self.p2.measure_angle(self.p1)
        a2 = self.p2.measure_angle(self.p3)
        measured_angle = (a1 - a2) % (2 * np.pi)
        return np.isclose(measured_angle, np.pi, rtol=1e-14)


class SymmetryConstraint(GeoCon):

    default_name = "SymCon-1"

    def __init__(self, p1: Point, p2: Point, p3: Point, p4: Point, name: str = None):
        self.p1 = p1  # Line start point
        self.p2 = p2  # Line end point
        self.p3 = p3  # Tool point
        self.p4 = p4  # Target point
        super().__init__(param=None, name=name, child_nodes=[self.p1, self.p2, self.p3, self.p4], kind="a4|d")

    @staticmethod
    def check_if_point_is_symmetric_target(p: Point):
        for geo_con in p.geo_cons:
            if isinstance(geo_con, SymmetryConstraint) and geo_con.child_nodes[-1] is p:
                return True
        return False

    def __repr__(self):
        return f"{self.__class__.__name__} {self.name()}"

    def get_dict_rep(self) -> dict:
        return {"p1": self.p1.name(), "p2": self.p2.name(), "p3": self.p3.name(), "p4": self.p4.name(),
                "constraint_type": self.__class__.__name__}

    def verify(self) -> bool:
        tool_angle = measure_rel_angle3(self.p3.x().value(), self.p3.y().value(),
                                        self.p2.x().value(), self.p2.y().value(),
                                        self.p1.x().value(), self.p1.y().value())
        target_angle = measure_rel_angle3(self.p1.x().value(), self.p1.y().value(),
                                          self.p2.x().value(), self.p2.y().value(),
                                          self.p4.x().value(), self.p4.y().value())
        if not np.isclose(tool_angle, target_angle, rtol=1e-14):
            return False

        tool_distance_to_line = measure_point_line_distance_unsigned(
            self.p1.x().value(), self.p1.y().value(),
            self.p2.x().value(), self.p2.y().value(),
            self.p3.x().value(), self.p3.y().value()
        )
        target_distance_to_line = measure_point_line_distance_unsigned(
            self.p1.x().value(), self.p1.y().value(),
            self.p2.x().value(), self.p2.y().value(),
            self.p4.x().value(), self.p4.y().value()
        )

        return np.isclose(tool_distance_to_line, target_distance_to_line, rtol=1e-14)


class Perp3Constraint(GeoCon):

    default_name = "Perp3Con-1"

    def __init__(self, p1: Point, p2: Point, p3: Point, name: str = None):
        self.p1 = p1
        self.p2 = p2
        self.p3 = p3
        super().__init__(param=None, name=name, child_nodes=[self.p1, self.p2, self.p3], kind="a3")

    def __repr__(self):
        return f"{self.__class__.__name__} {self.name()}"

    def get_dict_rep(self) -> dict:
        return {"p1": self.p1.name(), "p2": self.p2.name(), "p3": self.p3.name(),
                "constraint_type": self.__class__.__name__}

    def verify(self) -> bool:
        a1 = self.p2.measure_angle(self.p1)
        a2 = self.p2.measure_angle(self.p3)
        measured_angle = (a1 - a2) % (2 * np.pi)
        return np.isclose(measured_angle, 0.5 * np.pi, rtol=1e-14)


class RelAngle3Constraint(GeoCon):

    default_name = "RelAng3Con-1"

    def __init__(self, p1: Point, p2: Point, p3: Point, value: float or AngleParam = None, name: str = None):
        self.p1 = p1
        self.p2 = p2
        self.p3 = p3
        if value is None:
            args = []
            for point in [self.p1, self.p2, self.p3]:
                args.extend([point.x().value(), point.y().value()])
            value = measure_rel_angle3(*args)
        param = value if isinstance(value, Param) else AngleParam(value=value, name="Angle-1")
        super().__init__(param=param, name=name, child_nodes=[self.p1, self.p2, self.p3], kind="a3")

    def __repr__(self):
        return f"{self.__class__.__name__} {self.name()}<v={self.param().value()}>"

    def get_dict_rep(self) -> dict:
        return {"p1": self.p1.name(), "p2": self.p2.name(),
                "p3": self.p3.name(), "value": self.param().name(),
                "constraint_type": self.__class__.__name__}

    def verify(self) -> bool:
        a1 = self.p2.measure_angle(self.p1)
        a2 = self.p2.measure_angle(self.p3)
        measured_angle = (a1 - a2) % (2 * np.pi)
        return np.isclose(measured_angle, self.param().rad(), rtol=1e-14)


@dataclass
class CurvatureConstraintData:
    Lt1: float
    Lt2: float
    Lc1: float
    Lc2: float
    n1: int
    n2: int
    theta1: float
    theta2: float
    phi1: float
    phi2: float
    psi1: float
    psi2: float
    R1: float
    R2: float


class ROCurvatureConstraint(GeoCon):

    default_name = "ROCCon-1"

    def __init__(self, curve_joint: Point, value: float or LengthParam = None, name: str = None):
        if len(curve_joint.curves) != 2:
            raise ConstraintValidationError(f"There must be exactly two curves attached to the curve joint. Found "
                                            f"{len(curve_joint.curves)} curves")
        self.curve_joint = curve_joint
        self.curve_1 = curve_joint.curves[0]
        self.curve_2 = curve_joint.curves[1]
        self.curve_type_1 = self.curve_1.__class__.__name__
        self.curve_type_2 = self.curve_2.__class__.__name__

        if self.curve_type_1 == "Bezier":
            curve_joint_index_curve_1 = self.curve_1.point_sequence().points().index(curve_joint)
            self.curve_joint_index_curve_1 = -1 if curve_joint_index_curve_1 != 0 else 0
            self.g2_point_index_curve_1 = 2 if self.curve_joint_index_curve_1 == 0 else -3
            self.g1_point_index_curve_1 = 1 if self.g2_point_index_curve_1 == 2 else -2
            self.g1_point_curve_1 = self.curve_1.point_sequence().points()[self.g1_point_index_curve_1]
            self.g2_point_curve_1 = self.curve_1.point_sequence().points()[self.g2_point_index_curve_1]
        else:
            (curve_joint_index_curve_1, self.curve_joint_index_curve_1,
             self.g2_point_index_curve_1, self.g1_point_index_curve_1,
             self.g1_point_curve_1, self.g2_point_curve_1) = None, None, None, None, None, None

        if self.curve_type_2 == "Bezier":
            curve_joint_index_curve_2 = self.curve_2.point_sequence().points().index(curve_joint)
            self.curve_joint_index_curve_2 = -1 if curve_joint_index_curve_2 != 0 else 0
            self.g2_point_index_curve_2 = 2 if self.curve_joint_index_curve_2 == 0 else -3
            self.g1_point_index_curve_2 = 1 if self.g2_point_index_curve_2 == 2 else -2
            self.g1_point_curve_2 = self.curve_2.point_sequence().points()[self.g1_point_index_curve_2]
            self.g2_point_curve_2 = self.curve_2.point_sequence().points()[self.g2_point_index_curve_2]
        else:
            (curve_joint_index_curve_2, self.curve_joint_index_curve_2,
             self.g2_point_index_curve_2, self.g1_point_index_curve_2,
             self.g1_point_curve_2, self.g2_point_curve_2) = None, None, None, None, None, None

        points = [self.curve_joint]
        for point in [self.g2_point_curve_1, self.g1_point_curve_1, self.g1_point_curve_2, self.g2_point_curve_2]:
            if point is None:
                continue
            points.append(point)

        secondary_params = []
        if self.curve_type_1 == "Bezier":
            secondary_params.append(Param(self.curve_1.degree, "n"))
        if self.curve_type_2 == "Bezier":
            secondary_params.append(Param(self.curve_2.degree, "n"))

        if self.curve_type_1 == "PolyLine" and self.curve_type_2 == "PolyLine":
            raise ValueError("Cannot create radius of curvature constraint between two polylines")

        if self.curve_type_1 == "LineSegment" or self.curve_type_2 == "LineSegment":
            param = None
        elif self.curve_type_1 == "PolyLine" or (self.curve_type_1 == "Bezier" and (
                self.curve_1.t_start is not None or self.curve_1.t_end is not None)):
            data = self.curve_1.evaluate()
            if (np.isclose(data.xy[0, 0], self.curve_joint.x().value()) and
                np.isclose(data.xy[0, 1], self.curve_joint.y().value())):
                R = data.R[0]
            else:
                R = data.R[-1]

            param = LengthParam(value=R, name="ROC-1", enabled=False) if not isinstance(value, Param) else value
        elif self.curve_type_2 == "PolyLine" or (self.curve_type_2 == "Bezier" and (
                self.curve_2.t_start is not None or self.curve_2.t_end is not None)):
            data = self.curve_2.evaluate()
            if (np.isclose(data.xy[0, 0], self.curve_joint.x().value()) and
                    np.isclose(data.xy[0, 1], self.curve_joint.y().value())):
                R = data.R[0]
            else:
                R = data.R[-1]
            param = LengthParam(value=R, name="ROC-1", enabled=False) if not isinstance(value, Param) else value
        else:
            enabled = True
            if value is None:
                curvature_data = self.calculate_curvature_data(self.curve_joint)

                if not self.is_solving_allowed(self.g2_point_curve_1):
                    value = curvature_data.R1
                    enabled = False
                elif not self.is_solving_allowed(self.g2_point_curve_2):
                    value = curvature_data.R2
                    enabled = False
                else:
                    value = 0.5 * (curvature_data.R1 + curvature_data.R2)
            param = value if isinstance(value, Param) else LengthParam(value=value, name="ROC-1")
            param.set_enabled(enabled)

        super().__init__(param=param, child_nodes=points, kind="d", name=name,
                         secondary_params=secondary_params)

    @staticmethod
    def calculate_curvature_data(curve_joint):
        curve_1 = curve_joint.curves[0]
        if curve_1.__class__.__name__ == "Bezier":
            curve_joint_index_curve_1 = curve_1.point_sequence().points().index(curve_joint)
            curve_joint_index_curve_1 = -1 if curve_joint_index_curve_1 != 0 else 0
            g2_point_index_curve_1 = 2 if curve_joint_index_curve_1 == 0 else -3
            g1_point_index_curve_1 = 1 if g2_point_index_curve_1 == 2 else -2
            g1_point_curve_1 = curve_1.point_sequence().points()[g1_point_index_curve_1]
            g2_point_curve_1 = curve_1.point_sequence().points()[g2_point_index_curve_1]
            Lt1 = g1_point_curve_1.measure_distance(curve_joint)
            Lc1 = g1_point_curve_1.measure_distance(g2_point_curve_1)
            n1 = curve_1.degree
            phi1 = curve_joint.measure_angle(g1_point_curve_1)
            theta1 = g1_point_curve_1.measure_angle(g2_point_curve_1)
            psi1 = theta1 - phi1
            R1 = np.abs(np.true_divide((Lt1 * Lt1), (Lc1 * (1 - 1 / n1) * np.sin(psi1))))
        else:
            Lt1, Lc1, n1, theta1, phi1, psi1, R1 = None, None, None, None, None, None, None

        curve_2 = curve_joint.curves[1]
        if curve_2.__class__.__name__ == "Bezier":
            curve_joint_index_curve_2 = curve_2.point_sequence().points().index(curve_joint)
            curve_joint_index_curve_2 = -1 if curve_joint_index_curve_2 != 0 else 0
            g2_point_index_curve_2 = 2 if curve_joint_index_curve_2 == 0 else -3
            g1_point_index_curve_2 = 1 if g2_point_index_curve_2 == 2 else -2
            g1_point_curve_2 = curve_2.point_sequence().points()[g1_point_index_curve_2]
            g2_point_curve_2 = curve_2.point_sequence().points()[g2_point_index_curve_2]
            Lt2 = g1_point_curve_2.measure_distance(curve_joint)
            Lc2 = g1_point_curve_2.measure_distance(g2_point_curve_2)
            n2 = curve_2.degree
            phi2 = curve_joint.measure_angle(g1_point_curve_2)
            theta2 = g1_point_curve_2.measure_angle(g2_point_curve_2)
            psi2 = theta2 - phi2
            R2 = np.abs(np.true_divide((Lt2 * Lt2), (Lc2 * (1 - 1 / n2) * np.sin(psi2))))
        else:
            Lt2, Lc2, n2, theta2, phi2, psi2, R2 = None, None, None, None, None, None, None

        if curve_1.__class__.__name__ == "PolyLine":
            data = curve_1.evaluate()
            if (np.isclose(data.xy[0, 0], curve_joint.x().value()) and
                    np.isclose(data.xy[0, 1], curve_joint.y().value())):
                R1 = data.R[0]
            else:
                R1 = data.R[-1]
        if curve_2.__class__.__name__ == "PolyLine":
            data = curve_2.evaluate()
            if (np.isclose(data.xy[0, 0], curve_joint.x().value()) and
                    np.isclose(data.xy[0, 1], curve_joint.y().value())):
                R2 = data.R[0]
            else:
                R2 = data.R[-1]

        data = CurvatureConstraintData(Lt1=Lt1, Lt2=Lt2, Lc1=Lc1, Lc2=Lc2, n1=n1, n2=n2, theta1=theta1, theta2=theta2,
                                       phi1=phi1, phi2=phi2, psi1=psi1, psi2=psi2, R1=R1, R2=R2)
        return data

    @staticmethod
    def is_solving_allowed(g2_point: Point):
        symmetry_constraints = [geo_con for geo_con in g2_point.geo_cons if
                                isinstance(geo_con, SymmetryConstraint)]
        # Check if the point is the symmetry target point
        if any([symmetry_constraint.child_nodes[-1] is g2_point for symmetry_constraint in symmetry_constraints]):
            return False
        return True

    def __repr__(self):
        return f"{self.__class__.__name__} {self.name()} <C1={self.curve_1.name()}, C2={self.curve_2.name()}>"

    def get_dict_rep(self):
        value = self.param().name() if self.param() is not None else None
        return {"curve_joint": self.curve_joint.name(), "value": value,
                "constraint_type": self.__class__.__name__}

    def verify(self) -> bool:
        data = self.calculate_curvature_data(self.curve_joint)
        return np.isclose(data.R1, data.R2, rtol=1e-14)


class ConstraintValidationError(Exception):
    pass


class NoSolutionError(Exception):
    pass


class DuplicateConstraintError(Exception):
    pass


class MaxWeakConstraintAttemptsError(Exception):
    pass



import requests
from dataclasses import dataclass

from pymead.version import __version__


def get_current_pymead_version():
    """
    Gets the current pymead version from the top-level __version__

    Returns
    =======
    str
        Current version of pymead in raw string form
    """
    return "v" + __version__


def get_latest_pymead_release_version():
    """
    Gets the latest raw version string from pymead's GitHub release page (note that the order of the list generated
    by `response.json()` is in order from latest to oldest.

    Returns
    =======
    str
        Raw version string representing the latest version of pymead
    """
    response = requests.get("https://api.github.com/repos/mlau154/pymead/releases")
    latest_version_string = response.json()[0]["tag_name"]
    return latest_version_string


@dataclass
class VersionInfo:
    major: int
    minor: int
    patch: int
    pre_release_tag: str or None
    pre_release_id: int or None


def extract_version_info_from_string(version_string: str):
    """
    Takes a raw string of the form 'v[major].[minor].[patch]-[pre-release tag].[pre-release id]' and
    returns a namedtuple containing the fields identified in brackets.

    Parameters
    ==========
    version_string: str
        Raw version string

    Returns
    =======
    VersionInfo
    """
    v_split = version_string.split("-")

    if "v" in v_split[0]:
        major_minor_patch_str = v_split[0][1:]
    else:
        major_minor_patch_str = v_split[0][:]

    major = int(major_minor_patch_str.split(".")[0])
    minor = int(major_minor_patch_str.split(".")[1])
    patch = int(major_minor_patch_str.split(".")[2])

    pre_release_tag = None
    pre_release_id = None

    if len(v_split) > 1:
        pre_release_tag = v_split[1].split(".")[0]
        pre_release_id = int(v_split[1].split(".")[1])

    return VersionInfo(major=major, minor=minor, patch=patch,
                       pre_release_tag=pre_release_tag, pre_release_id=pre_release_id)


def compare_versions(version_1: VersionInfo or str, version_2: VersionInfo or str):
    """
    Compares two versions. If the two versions are equal, return 2. If version 2 is newer than version 1, return 1.
    If version 2 is older than version 1, return 0.

    Parameters
    ==========
    version_1: VersionInfo or str
        First version to compare

    version_2: VersionInfo or str
        Second version to compare

    Returns
    =======
    int
        Flag representing the result of the version comparison. 0: version 2 is older than version 1. 1: version 2
        is newer than version 1. 2: the versions match.
    """

    if not isinstance(version_1, VersionInfo):
        version_1 = extract_version_info_from_string(version_1)
    if not isinstance(version_2, VersionInfo):
        version_2 = extract_version_info_from_string(version_2)

    def compare_values(v1: int, v2: int):
        if v2 == v1:
            return None
        elif v2 > v1:
            return 1
        else:
            return 0

    # First, compare the major version
    major_comparison = compare_values(version_1.major, version_2.major)
    if major_comparison is not None:
        return major_comparison

    # Next, compare the minor version
    minor_comparison = compare_values(version_1.minor, version_2.minor)
    if minor_comparison is not None:
        return minor_comparison

    # Then, compare the patch version
    patch_comparison = compare_values(version_1.patch, version_2.patch)
    if patch_comparison is not None:
        return patch_comparison

    # If the major, minor, and patch versions all match but only one version has a pre-release tag, use this information
    # directly to determine the result of the comparison
    if version_1.pre_release_tag is None and version_2.pre_release_tag is not None:
        return 0
    if version_1.pre_release_tag is not None and version_2.pre_release_tag is None:
        return 1

    # If the pre-release tags do not match, use this information directly to determine the result of the comparison
    if version_1.pre_release_tag != version_2.pre_release_tag:
        if version_1.pre_release_tag == "alpha" and version_2.pre_release_tag == "beta":
            return 1
        elif version_1.pre_release_tag == "beta" and version_2.pre_release_tag == "alpha":
            return 0
        else:
            raise ValueError(f"Found invalid pre_release_tag {version_1.pre_release_tag = }, "
                             f"{version_2.pre_release_tag = }. Valid pre-release tag names are 'alpha' and 'beta'.")

    # Perform the pre-release id comparison
    pre_release_id_comparison = compare_values(version_1.pre_release_id, version_2.pre_release_id)
    if pre_release_id_comparison is not None:
        return pre_release_id_comparison

    # If the function has not yet returned a value, this means that the version strings match
    return 2


def using_latest(current_version_string: str or None = None):
    """
    Compares the current version string to the latest version string. If the latest version string is > the current
    version string, return False, else return True. If the current version string is None, pull the version from
    `pymead.version.__version__`.

    Parameters
    ==========
    current_version_string: str or None
        The version of pymead in use. Default: `None`

    Returns
    =======
    typing.Tuple[bool, str, str]
        First element: True if using the latest version, False otherwise. Second element: string representing the latest
        version of pymead. Third element: string representing the current version of pymead.
    """
    if current_version_string is None:
        current_version_string = get_current_pymead_version()

    current_version = extract_version_info_from_string(current_version_string)

    latest_version_string = get_latest_pymead_release_version()
    latest_version = extract_version_info_from_string(latest_version_string)

    return bool(compare_versions(latest_version, current_version)), latest_version_string, current_version_string



"""
Inspired by https://forum.qt.io/topic/110138/show-qprogressbar-with-computationally-heavy-background-process/2
"""

import multiprocessing as mp
from multiprocessing import active_children

from PyQt5.QtCore import QObject, QRunnable, QThreadPool, pyqtSignal

from pymead.utils.pymead_mp import kill_child_processes


class ProgressEmitter(QRunnable):
    """Listens to status of process"""

    class ProgressSignals(QObject):
        progress = pyqtSignal(str, object)
        finished = pyqtSignal(bool)

    def __init__(self, conn, process):
        super().__init__()
        self.conn = conn
        self.process = process
        self.signals = ProgressEmitter.ProgressSignals()
        self.running = True

    def run(self):
        while self.running:
            try:
                if self.conn.poll():
                    try:
                        status, data = self.conn.recv()
                        self.signals.progress.emit(status, data)
                        if status in ["finished"]:
                            self.signals.finished.emit(True)
                            self.running = False
                    except EOFError:
                        self.signals.finished.emit(True)
                        self.running = False
                elif not self.process.is_alive():
                    self.signals.progress.emit("terminated", None)
                    self.running = False
            except (BrokenPipeError, OSError, EOFError):
                self.signals.finished.emit(False)
                self.running = False


class CPUBoundProcess(QObject):

    def __init__(self, operation, args=(), parent=None):
        super().__init__(parent=parent)

        # Connection pipeline
        self.parent_conn, self.child_conn = mp.Pipe(duplex=True)

        # Process creation
        args = (self.child_conn, *args)
        self.process = mp.Process(target=operation, args=args)

        # Status emitter
        self.progress_emitter = ProgressEmitter(self.parent_conn, self.process)
        self.thread_pool = QThreadPool()

    def start(self):
        self.process.start()
        self.thread_pool.start(self.progress_emitter)

    def terminate(self):
        self.progress_emitter.running = False
        self.parent_conn.close()
        self.child_conn.close()
        for child in active_children():
            kill_child_processes(child.pid)
        self.process.terminate()



import sys
from abc import abstractmethod
from copy import deepcopy

from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot, QRegularExpression, QStringListModel
from PyQt5.QtGui import QValidator, QBrush, QColor
from PyQt5.QtWidgets import QTreeWidget, QTreeWidgetItem, QPushButton, QHBoxLayout, QHeaderView, QDialog, QGridLayout, \
    QDoubleSpinBox, QLineEdit, QLabel, QMenu, QAbstractItemView, QTreeWidgetItemIterator, QWidget, QCompleter

from pymead.core.airfoil import Airfoil
from pymead.core.bezier import Bezier
from pymead.core.constraints import *
from pymead.core.geometry_collection import GeometryCollection
from pymead.core.line import LineSegment, PolyLine
from pymead.core.mea import MEA
from pymead.core.param import Param, DesVar, LengthParam, AngleParam, LengthDesVar, AngleDesVar, EquationCompileError
from pymead.core.point import Point
from pymead.core.pymead_obj import PymeadObj
from pymead.gui.dialogs import PymeadDialog


class HeaderButtonRow(QHeaderView):
    sigExpandPressed = pyqtSignal()
    sigCollapsePressed = pyqtSignal()

    def __init__(self, parent):
        super().__init__(Qt.Horizontal, parent)
        self.lay = QHBoxLayout(self)
        self.expandButton = QPushButton("Expand All", self)
        self.collapseButton = QPushButton("Collapse All", self)
        self.expandButton.clicked.connect(self.expandButtonPressed)
        self.collapseButton.clicked.connect(self.collapseButtonPressed)
        self.lay.addWidget(self.expandButton)
        self.lay.addWidget(self.collapseButton)
        self.setLayout(self.lay)
        self.setFixedHeight(40)

    @pyqtSlot()
    def expandButtonPressed(self):
        self.sigExpandPressed.emit()

    @pyqtSlot()
    def collapseButtonPressed(self):
        self.sigCollapsePressed.emit()


class ValueSpin(QDoubleSpinBox):
    def __init__(self, parent, param: Param):
        super().__init__(parent)
        self.pymead_obj = param
        self.setMaximumWidth(150)
        self.setDecimals(6)
        self.setSingleStep(0.01)
        if isinstance(param, LengthParam) or isinstance(param, AngleParam):
            self.setSuffix(f" {param.unit()}")
        self.param = param
        if self.param.lower() is not None:
            self.setMinimum(self.param.lower())
        else:
            # if isinstance(self.param, LengthParam) or isinstance(self.param, AngleParam):
            #     self.setMinimum(0.0)
            # else:
            self.setMinimum(-1.0e9)
        if self.param.upper() is not None:
            self.setMaximum(self.param.upper())
        else:
            # if isinstance(self.param, AngleParam):
            #     self.setMaximum(UNITS.convert_angle_to_base(2 * np.pi, self.param.unit()))
            # else:
            self.setMaximum(1.0e9)
        self.setValue(self.param.value())
        self.valueChanged.connect(self.onValueChanged)
        self.setEnabled(self.param.enabled())

    # def setValue(self, val):
    #
    #     print(f"{val = }")
    #
    #     if isinstance(self.param, LengthParam) and self.param.point is None and val < 0.0:
    #         return
    #     elif isinstance(self.param, AngleParam):
    #         val = val % (2 * np.pi)
    #
    #     super().setValue(val)

    # def validate(self, inp, pos):
    #     if not hasattr(self, "param"):
    #         return QValidator.Acceptable
    #     elif isinstance(self.param, LengthParam) or isinstance(self.param, AngleParam) and len(inp.split()) > 1:
    #
    #         print(f"{inp = }, {pos = }")
    #         val = float(inp.split()[0])
    #         print(f"{val = }")
    #
    #         if isinstance(self.param, LengthParam) and val > 0.0:
    #             return QValidator.Acceptable
    #
    #         if isinstance(self.param, AngleParam) and 0.0 <= val < UNITS.convert_angle_to_base(2 * np.pi,
    #                                                                                            self.param.unit()):
    #             return QValidator.Acceptable
    #
    #         return QValidator.Intermediate
    #     else:
    #         return QValidator.Acceptable
    #
    # def fixup(self, s):
    #     suffix = None
    #     s_split = s.split()
    #     number = float(s_split[0])
    #     if len(s_split) > 1:
    #         suffix = s_split[1]
    #     print(f"{s = }")
    #
    #     if (isinstance(self.param, LengthParam) or isinstance(self.param, AngleParam)) and number < 0.0:
    #         return f"0.0 {suffix}"
    #
    #     if isinstance(self.param, AngleParam) and number > UNITS.convert_angle_from_base(2 * np.pi, self.param.unit()):
    #         return f"{UNITS.convert_angle_from_base(2 * np.pi)} {suffix}"

    def onValueChanged(self, value: float):
        if self.param.point is None:
            self.param.set_value(value)
        else:
            if self.param is self.param.point.x():
                self.param.point.request_move(value, self.param.point.y().value())
            elif self.param is self.param.point.y():
                self.param.point.request_move(self.param.point.x().value(), value)
        self.setValue(self.param.value())


class NameValidator(QValidator):

    def __init__(self, parent, tree, sub_container: str):
        super().__init__(parent)
        self.geo_col = tree.geo_col
        self.sub_container = sub_container
        self.regex = QRegularExpression("^[a-z-A-Z_0-9]+$")

    def validate(self, a0, a1):
        if a0 in self.geo_col.container()[self.sub_container].keys():
            return QValidator.Invalid, a0, a1

        if not self.regex.match(a0).hasMatch():
            return QValidator.Invalid, a0, a1

        return QValidator.Acceptable, a0, a1

    def fixup(self, a0):
        pass


class NameEdit(QLineEdit):
    def __init__(self, parent, pymead_obj: PymeadObj, tree):
        super().__init__(parent)
        self.pymead_obj = pymead_obj
        self.tree = tree

        validator = NameValidator(self, tree, sub_container=pymead_obj.sub_container)
        self.setValidator(validator)
        self.setText(self.pymead_obj.name())
        self.textChanged.connect(self.onTextChanged)

    def onTextChanged(self, name: str):
        self.pymead_obj.set_name(name)


class LowerSpin(QDoubleSpinBox):
    def __init__(self, parent, param: Param):
        super().__init__(parent)
        self.setDecimals(6)
        self.setSingleStep(0.01)
        self.param = param
        self.setMinimum(-1e9)
        self.setMaximum(1e9)
        self.setValue(self.param.lower())
        self.valueChanged.connect(self.onValueChanged)

    def onValueChanged(self, lower: float):
        self.param.set_lower(lower)
        self.setValue(self.param.lower())


class UpperSpin(QDoubleSpinBox):
    def __init__(self, parent, param: Param):
        super().__init__(parent)
        self.setDecimals(6)
        self.setSingleStep(0.01)
        self.param = param
        self.setMinimum(-1e9)
        self.setMaximum(1e9)
        self.setValue(self.param.upper())
        self.valueChanged.connect(self.onValueChanged)

    def onValueChanged(self, upper: float):
        self.param.set_upper(upper)


class TreeButton(QPushButton):
    sigNameChanged = pyqtSignal(str, object)

    def __init__(self, pymead_obj: PymeadObj, tree, top_level: bool = False):
        label = "Edit" if top_level else pymead_obj.name()
        self.top_level = top_level
        super().__init__(label)
        self.setMaximumWidth(150)
        self.pymead_obj = pymead_obj
        self.tree = tree
        self.dialog = None
        self.clicked.connect(self.onClicked)

    def onClicked(self):
        self.dialog = self.createDialog()
        if self.dialog.exec_():
            pass
        self.dialog = None
        self.tree.geo_col.clear_selected_objects()

    def onNameChange(self, name: str):
        if self.dialog is not None:
            self.dialog.setWindowTitle(f"{name}")
        if self.top_level:
            self.pymead_obj.tree_item.setText(0, name)
        else:
            self.setText(name)
        self.sigNameChanged.emit(name, self.pymead_obj)

    def createDialog(self):
        widget = QWidget()
        theme = self.tree.gui_obj.themes[self.tree.gui_obj.current_theme]
        dialog = PymeadDialog(self, window_title=f"{self.pymead_obj.name()}", widget=widget, theme=theme)
        layout = QGridLayout()
        widget.setLayout(layout)
        self.modifyDialogInternals(dialog, layout)
        return dialog

    @abstractmethod
    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        pass


class Completer(QCompleter):
    """
    From https://gitter.im/baudren/NoteOrganiser?at=55afbefdcce129d570a3c188
    """

    def __init__(self, model=None, parent=None):
        if model is None:
            super().__init__(parent)
        else:
            super().__init__(model, parent)

        self.setCaseSensitivity(Qt.CaseInsensitive)
        self.setCompletionMode(QCompleter.PopupCompletion)
        self.setWrapAround(False)

    # Add texts instead of replace
    def pathFromIndex(self, index):
        path = QCompleter.pathFromIndex(self, index)

        print(f"At the start, {path = }")

        lst = str(self.widget().text()).split('$')

        if len(lst) > 1:
            # path = "$".join(lst[:-1]) + path
            path = "$" + path

        print(f"{lst = }, {path = }, {self.widget().text() = }")

        return path

    # Add operator to separate between texts
    def splitPath(self, path):
        for ch in [" ", "+", " - ", "*", "/", "(", "$"]:
            path = str(path.split(ch)[-1])
            print(f"{path = }")
        return [path]


class ParamEquationEdit(QLineEdit):
    def __init__(self, param: Param, parent=None):
        super().__init__(parent=parent)
        self.param = param
        # completer = Completer([p.name() for p in self.param.param_graph.param_list], self)
        # self.setCompleter(completer)  # TODO: fix autocomplete (removed for now)
        self.editingFinished.connect(self.updateEquation)
        self.setPlaceholderText("$")
        if self.param.equation_str is not None:
            self.setText(self.param.equation_str)

    def updateEquation(self):
        try:
            self.param.update_equation(self.text())
        except EquationCompileError as e:
            if self.param.geo_col and self.param.geo_col.gui_obj:
                self.param.geo_col.gui_obj.disp_message_box(str(e))
                return
            else:
                raise e


class ParamButton(TreeButton):
    sigValueChanged = pyqtSignal(float)  # value

    def __init__(self, param: Param, tree, name_editable: bool = True, top_level: bool = False):
        super().__init__(pymead_obj=param, tree=tree, top_level=top_level)
        self.name_editable = name_editable
        self.param = param

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        value_label = QLabel("Value", self)
        value_spin = ValueSpin(self, self.param)
        value_spin.valueChanged.connect(self.onValueChange)
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.param, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        if not self.name_editable:
            name_edit.setReadOnly(True)
        layout.addWidget(value_label, 0, 0)
        layout.addWidget(value_spin, 0, 1)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)
        if self.param.lower() is not None:
            lower_label = QLabel("Lower Bound", self)
            lower_spin = LowerSpin(self, self.param)
            row_count = layout.rowCount()
            layout.addWidget(lower_label, row_count, 0)
            layout.addWidget(lower_spin, row_count, 1)
        if self.param.upper() is not None:
            upper_label = QLabel("Upper Bound", self)
            upper_spin = UpperSpin(self, self.param)
            row_count = layout.rowCount()
            layout.addWidget(upper_label, row_count, 0)
            layout.addWidget(upper_spin, row_count, 1)
        row_count = layout.rowCount()
        layout.addWidget(QLabel("Equation"), row_count, 0)
        layout.addWidget(ParamEquationEdit(param=self.param, parent=self), row_count, 1)

    def onValueChange(self, value: float):
        self.sigValueChanged.emit(value)


class LengthParamButton(ParamButton):
    pass


class AngleParamButton(ParamButton):
    pass


class DesVarButton(TreeButton):
    sigValueChanged = pyqtSignal(float)  # value

    def __init__(self, desvar: DesVar, tree, name_editable: bool = True, top_level: bool = False):
        super().__init__(pymead_obj=desvar, tree=tree, top_level=top_level)
        self.name_editable = name_editable
        self.desvar = desvar

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        value_label = QLabel("Value", self)
        value_spin = ValueSpin(self, self.desvar)
        value_spin.valueChanged.connect(self.onValueChange)
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.desvar, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        if not self.name_editable:
            name_edit.setReadOnly(True)
        layout.addWidget(value_label, 0, 0)
        layout.addWidget(value_spin, 0, 1)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)
        lower_label = QLabel("Lower Bound", self)
        lower_spin = LowerSpin(self, self.desvar)
        row_count = layout.rowCount()
        layout.addWidget(lower_label, row_count, 0)
        layout.addWidget(lower_spin, row_count, 1)
        upper_label = QLabel("Upper Bound", self)
        upper_spin = UpperSpin(self, self.desvar)
        row_count = layout.rowCount()
        layout.addWidget(upper_label, row_count, 0)
        layout.addWidget(upper_spin, row_count, 1)
        row_count = layout.rowCount()
        layout.addWidget(QLabel("Equation"), row_count, 0)
        layout.addWidget(ParamEquationEdit(param=self.desvar, parent=self), row_count, 1)

    def onValueChange(self, value: float):
        self.sigValueChanged.emit(value)


class LengthDesVarButton(DesVarButton):
    pass


class AngleDesVarButton(DesVarButton):
    pass


class PointButton(TreeButton):

    def __init__(self, point: Point, tree, top_level: bool = False):
        super().__init__(pymead_obj=point, tree=tree, top_level=top_level)
        self.point = point
        self.x_button = None
        self.y_button = None

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.point, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        x_label = QLabel("x", self)
        self.x_button = ParamButton(self.point.x(), self.tree, name_editable=False)
        self.x_button.sigValueChanged.connect(self.onXChanged)
        y_label = QLabel("y", self)
        self.y_button = ParamButton(self.point.y(), self.tree, name_editable=False)
        self.y_button.sigValueChanged.connect(self.onYChanged)
        layout.addWidget(name_label, 0, 0)
        layout.addWidget(name_edit, 0, 1)
        layout.addWidget(x_label, 1, 0)
        layout.addWidget(self.x_button, 1, 1)
        layout.addWidget(y_label, 2, 0)
        layout.addWidget(self.y_button, 2, 1)

    def onXChanged(self, x: float):
        self.point.request_move(x, self.point.y().value())

    def onYChanged(self, y: float):
        self.point.request_move(self.point.x().value(), y)

    def onNameChange(self, name: str):
        self.x_button.setText(f"{name}.x")
        self.y_button.setText(f"{name}.y")
        super().onNameChange(name=name)

    def enterEvent(self, a0):
        if self.top_level:
            return
        if self.pymead_obj.tree_item.hoverable:
            self.tree.geo_col.hover_enter_obj(self.pymead_obj)

    def leaveEvent(self, a0):
        if self.top_level:
            return
        if self.pymead_obj.tree_item.hoverable:
            self.tree.geo_col.hover_leave_obj(self.pymead_obj)


class BezierButton(TreeButton):

    def __init__(self, bezier: Bezier, tree, top_level: bool = False):
        super().__init__(pymead_obj=bezier, tree=tree, top_level=top_level)
        self.bezier = bezier

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.bezier, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)
        for point in self.bezier.point_sequence().points():
            point_button = PointButton(point, self.tree)
            point_button.sigNameChanged.connect(self.onPointNameChange)
            layout.addWidget(point_button, layout.rowCount(), 0)

    def onPointNameChange(self, name: str, point: Point):
        if point.tree_item is not None:
            self.tree.itemWidget(point.tree_item, 0).setText(name)

    def enterEvent(self, a0):
        self.bezier.canvas_item.setCurveStyle("hovered")

    def leaveEvent(self, a0):
        self.bezier.canvas_item.setCurveStyle("default")


class LineSegmentButton(TreeButton):

    def __init__(self, line: LineSegment, tree, top_level: bool = False):
        super().__init__(pymead_obj=line, tree=tree, top_level=top_level)
        self.line = line

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.line, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)
        for point in self.line.point_sequence().points():
            point_button = PointButton(point, self.tree)
            point_button.sigNameChanged.connect(self.onPointNameChange)
            layout.addWidget(point_button, layout.rowCount(), 0)

    def onPointNameChange(self, name: str, point: Point):
        if point.tree_item is not None:
            self.tree.itemWidget(point.tree_item, 0).setText(name)

    def enterEvent(self, a0):
        self.line.canvas_item.setCurveStyle("hovered")

    def leaveEvent(self, a0):
        self.line.canvas_item.setCurveStyle("default")


class AirfoilButton(TreeButton):
    def __init__(self, airfoil: Airfoil, tree, top_level: bool = False):
        super().__init__(pymead_obj=airfoil, tree=tree, top_level=top_level)
        self.airfoil = airfoil

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.airfoil, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)
        labels = ["Leading Edge", "Trailing Edge", "Upper Surface End", "Lower Surface End"]
        points = [self.airfoil.leading_edge, self.airfoil.trailing_edge, self.airfoil.upper_surf_end,
                  self.airfoil.lower_surf_end]

        for label, point in zip(labels, points):
            q_label = QLabel(label)
            point_button = PointButton(point, self.tree)
            point_button.sigNameChanged.connect(self.onPointNameChange)
            row_count = layout.rowCount()
            layout.addWidget(q_label, row_count, 0)
            layout.addWidget(point_button, row_count, 1)

    def onPointNameChange(self, name: str, point: Point):
        if point.tree_item is not None:
            self.tree.itemWidget(point.tree_item, 0).setText(name)


class PolyLineButton(TreeButton):
    def __init__(self, polyline: PolyLine, tree, top_level: bool = False):
        super().__init__(pymead_obj=polyline, tree=tree, top_level=top_level)
        self.polyline = polyline

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.polyline, self.tree)
        name_edit.setEnabled(False)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)


class MEAButton(TreeButton):
    def __init__(self, mea: MEA, tree, top_level: bool = False):
        super().__init__(pymead_obj=mea, tree=tree, top_level=top_level)
        self.mea = mea

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.mea, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)

        for airfoil in self.mea.airfoils:
            airfoil_button = AirfoilButton(airfoil, self.tree)
            airfoil_button.sigNameChanged.connect(self.onAirfoilNameChange)
            layout.addWidget(airfoil_button, layout.rowCount(), 0)

    def onAirfoilNameChange(self, name: str, airfoil: Airfoil):
        if airfoil.tree_item is not None:
            self.tree.itemWidget(airfoil.tree_item, 0).setText(name)


class DistanceConstraintButton(TreeButton):

    def __init__(self, distance_constraint: DistanceConstraint, tree, top_level: bool = False):
        super().__init__(pymead_obj=distance_constraint, tree=tree, top_level=top_level)
        self.distance_constraint = distance_constraint

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.distance_constraint, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)
        labels = ["Start Point", "End Point"]
        points = [self.distance_constraint.p1, self.distance_constraint.p2]
        for label, point in zip(labels, points):
            point_button = PointButton(point, self.tree)
            point_button.sigNameChanged.connect(self.onPointNameChange)
            q_label = QLabel(label, self)
            row_count = layout.rowCount()
            layout.addWidget(q_label, row_count, 0)
            layout.addWidget(point_button, row_count, 1)
        row_count = layout.rowCount()
        layout.addWidget(QLabel("Length Param", self), row_count, 0)
        layout.addWidget(LengthParamButton(self.distance_constraint.param(), self.tree), row_count, 1)

    def onPointNameChange(self, name: str, point: Point):
        if point.tree_item is not None:
            self.tree.itemWidget(point.tree_item, 0).setText(name)


class RelAngle3ConstraintButton(TreeButton):

    def __init__(self, rel_angle_constraint: RelAngle3Constraint, tree, top_level: bool = False):
        super().__init__(pymead_obj=rel_angle_constraint, tree=tree, top_level=top_level)
        self.rel_angle_constraint = rel_angle_constraint

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.rel_angle_constraint, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)

        row_count = layout.rowCount()
        layout.addWidget(QLabel("Angle Param", self), row_count, 0)
        layout.addWidget(AngleParamButton(self.rel_angle_constraint.param(), self.tree), row_count, 1)

        labels = ["Start", "Vertex", "End"]
        points = [self.rel_angle_constraint.p1, self.rel_angle_constraint.p2,
                  self.rel_angle_constraint.p3]
        for label, point in zip(labels, points):
            point_button = PointButton(point, self.tree)
            point_button.sigNameChanged.connect(self.onPointNameChange)
            q_label = QLabel(label, self)
            row_count = layout.rowCount()
            layout.addWidget(q_label, row_count, 0)
            layout.addWidget(point_button, row_count, 1)

    def onPointNameChange(self, name: str, point: Point):
        if point.tree_item is not None:
            self.tree.itemWidget(point.tree_item, 0).setText(name)


class Perp3ConstraintButton(TreeButton):

    def __init__(self, perp3_constraint: Perp3Constraint, tree, top_level: bool = False):
        super().__init__(pymead_obj=perp3_constraint, tree=tree, top_level=top_level)
        self.perp3_constraint = perp3_constraint

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.perp3_constraint, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)

        labels = ["Start", "Vertex", "End"]
        points = [self.perp3_constraint.p1, self.perp3_constraint.p2,
                  self.perp3_constraint.p3]
        for label, point in zip(labels, points):
            point_button = PointButton(point, self.tree)
            point_button.sigNameChanged.connect(self.onPointNameChange)
            q_label = QLabel(label, self)
            row_count = layout.rowCount()
            layout.addWidget(q_label, row_count, 0)
            layout.addWidget(point_button, row_count, 1)

    def onPointNameChange(self, name: str, point: Point):
        if point.tree_item is not None:
            self.tree.itemWidget(point.tree_item, 0).setText(name)


class AntiParallel3ConstraintButton(TreeButton):

    def __init__(self, antiparallel3_constraint: AntiParallel3Constraint, tree, top_level: bool = False):
        super().__init__(pymead_obj=antiparallel3_constraint, tree=tree, top_level=top_level)
        self.antiparallel3_constraint = antiparallel3_constraint

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.antiparallel3_constraint, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)

        labels = ["Start", "Vertex", "End"]
        points = [self.antiparallel3_constraint.p1, self.antiparallel3_constraint.p2,
                  self.antiparallel3_constraint.p3]
        for label, point in zip(labels, points):
            point_button = PointButton(point, self.tree)
            point_button.sigNameChanged.connect(self.onPointNameChange)
            q_label = QLabel(label, self)
            row_count = layout.rowCount()
            layout.addWidget(q_label, row_count, 0)
            layout.addWidget(point_button, row_count, 1)

    def onPointNameChange(self, name: str, point: Point):
        if point.tree_item is not None:
            self.tree.itemWidget(point.tree_item, 0).setText(name)


class SymmetryConstraintButton(TreeButton):

    def __init__(self, symmetry_constraint: SymmetryConstraint, tree, top_level: bool = False):
        super().__init__(pymead_obj=symmetry_constraint, tree=tree, top_level=top_level)
        self.symmetry_constraint = symmetry_constraint

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.symmetry_constraint, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)

        labels = ["Mirror Start", "Mirror End", "Tool Point", "Target Point"]
        points = [self.symmetry_constraint.p1, self.symmetry_constraint.p2,
                  self.symmetry_constraint.p3, self.symmetry_constraint.p4]
        for label, point in zip(labels, points):
            point_button = PointButton(point, self.tree)
            point_button.sigNameChanged.connect(self.onPointNameChange)
            q_label = QLabel(label, self)
            row_count = layout.rowCount()
            layout.addWidget(q_label, row_count, 0)
            layout.addWidget(point_button, row_count, 1)

    def onPointNameChange(self, name: str, point: Point):
        if point.tree_item is not None:
            self.tree.itemWidget(point.tree_item, 0).setText(name)


class ROCurvatureConstraintButton(TreeButton):

    def __init__(self, curvature_constraint: ROCurvatureConstraint, tree, top_level: bool = False):
        super().__init__(pymead_obj=curvature_constraint, tree=tree, top_level=top_level)
        self.curvature_constraint = curvature_constraint

    def modifyDialogInternals(self, dialog: QDialog, layout: QGridLayout) -> None:
        name_label = QLabel("Name", self)
        name_edit = NameEdit(self, self.curvature_constraint, self.tree)
        name_edit.textChanged.connect(self.onNameChange)
        layout.addWidget(name_label, 1, 0)
        layout.addWidget(name_edit, 1, 1)

        row_count = layout.rowCount()
        layout.addWidget(QLabel("Radius of Curvature", self), row_count, 0)
        layout.addWidget(LengthParamButton(self.curvature_constraint.param(), self.tree), row_count, 1)

        labels = ["Curve 1 G2 Point", "Curve 1 G1 Point", "Curve Joint", "Curve 2 G1 Point", "Curve 2 G2 Point"]
        points = [self.curvature_constraint.g2_point_curve_1, self.curvature_constraint.g1_point_curve_1,
                  self.curvature_constraint.curve_joint,
                  self.curvature_constraint.g1_point_curve_2, self.curvature_constraint.g2_point_curve_2]
        for label, point in zip(labels, points):
            point_button = PointButton(point, self.tree)
            point_button.sigNameChanged.connect(self.onPointNameChange)
            q_label = QLabel(label, self)
            row_count = layout.rowCount()
            layout.addWidget(q_label, row_count, 0)
            layout.addWidget(point_button, row_count, 1)

    def onPointNameChange(self, name: str, point: Point):
        if point.tree_item is not None:
            self.tree.itemWidget(point.tree_item, 0).setText(name)


class PymeadTreeWidgetItem(QTreeWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hoverable = True


class ParameterTree(QTreeWidget):
    def __init__(self, geo_col: GeometryCollection, parent, gui_obj):
        super().__init__(parent)

        # Exchange references with the geometry collection
        self.geo_col = geo_col
        self.geo_col.tree = self

        # Access to GUI object
        self.gui_obj = gui_obj

        # Single column for the tree
        self.setColumnCount(2)

        # Aliases (suitable for display) for the sub-containers
        self.container_titles = {
            "desvar": "Design Variables",
            "params": "Parameters",
            "points": "Points",
            "lines": "Lines",
            "bezier": "Bézier Curves",
            "airfoils": "Airfoils",
            "polylines": "Polylines",
            "mea": "Multi-Element Airfoils",
            "geocon": "Geometric Constraints",
            "dims": "Dimensions"
        }
        self.inverse_mapped_containers = {v: k for k, v in self.container_titles.items()}

        # Set the top-level items (sub_containers)
        self.items = None
        self.topLevelDict = None
        self.addContainers()

        # Make the header
        self.setHeaderLabel("")
        self.headerRow = HeaderButtonRow(self)
        self.headerRow.sigExpandPressed.connect(self.onExpandPressed)
        self.headerRow.sigCollapsePressed.connect(self.onCollapsePressed)
        self.setHeader(self.headerRow)

        # Set the tree widget geometry
        self.setMinimumWidth(300)
        self.header().setSectionResizeMode(0, QHeaderView.Stretch)
        self.setColumnWidth(1, 120)

        # Set the tree to be expanded by default
        self.expandAll()

        # Set the selection mode to extended. This allows the user to perform the usual operations of Shift-Click,
        # Ctrl-Click, or drag to select multiple tree items at once
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)

        # Item selection changed connection
        self.previous_items = None
        self.itemSelectionChanged.connect(self.onItemSelectionChanged)

        # Allow mouse tracking so we can implement a hover method
        self.setMouseTracking(True)

        # Set double-click behavior
        self.doubleClicked.connect(self.onDoubleClick)

        # Previous item hovered
        self.previous_item_hovered = None

    def getPymeadObjFromItem(self, item: QTreeWidgetItem):
        pymead_obj_name = item.data(0, Qt.DisplayRole)
        pymead_obj_parent_tree_item = item.parent()
        if pymead_obj_parent_tree_item is None:
            return
        pymead_obj_parent_name = pymead_obj_parent_tree_item.data(0, Qt.DisplayRole)
        pymead_obj_sub_container = self.inverse_mapped_containers[pymead_obj_parent_name]
        try:
            pymead_obj = self.geo_col.container()[pymead_obj_sub_container][pymead_obj_name]
            return pymead_obj
        except KeyError:
            return

    def onDoubleClick(self, index):
        item = self.itemFromIndex(index)
        pymead_obj = self.getPymeadObjFromItem(item)

        if pymead_obj is None:
            return

        # Create a ghost TreeButton based on the pymead_obj and click it
        button_args = (pymead_obj, self)
        button = getattr(sys.modules[__name__], f"{type(pymead_obj).__name__}Button")(*button_args, top_level=True)
        button.setParent(self)
        button.onClicked()

    def onItemSelectionChanged(self):
        if self.previous_items is not None:
            for item in self.previous_items:
                pymead_obj = self.getPymeadObjFromItem(item)
                if pymead_obj is not None and item not in self.selectedItems():
                    self.geo_col.deselect_object(pymead_obj)

        for item in self.selectedItems():
            pymead_obj = self.getPymeadObjFromItem(item)
            if pymead_obj is not None:
                self.geo_col.select_object(pymead_obj)

        self.previous_items = self.selectedItems()

    def addContainers(self):
        self.items = [PymeadTreeWidgetItem(
            None, [f"{self.container_titles[k]}"]) for k in self.geo_col.container().keys()]
        self.topLevelDict = {k: i for i, k in enumerate(self.geo_col.container().keys())}
        self.insertTopLevelItems(0, self.items)

        # Sort the items in ascending order (A to Z)
        self.sortItems(0, Qt.SortOrder.AscendingOrder)

    def mouseMoveEvent(self, event):
        """
        Since the QTreeWidget does not emit a Hover signal, we effectively create one here by tracking the position
        of the mouse when it is inside the Parameter Tree and check whether there is a PymeadTreeWidgetItem under
        the mouse.
        """
        # Tracks the tree widget for a hover event, since a hover signal is not implemented in QTreeWidget
        tree_item = self.itemAt(event.x(), event.y())

        # Hover leave
        if (self.previous_item_hovered is not None and tree_item is not self.previous_item_hovered and
                self.previous_item_hovered.hoverable):
            pymead_obj = self.getPymeadObjFromItem(self.previous_item_hovered)
            if pymead_obj is not None:
                self.geo_col.hover_leave_obj(pymead_obj)
            else:
                self.setItemStyle(self.previous_item_hovered, "default")

        if not isinstance(tree_item, PymeadTreeWidgetItem):
            self.previous_item_hovered = tree_item
            return

        if tree_item.hoverable and tree_item is not None:
            # Hover enter
            pymead_obj = self.getPymeadObjFromItem(tree_item)
            if pymead_obj is not None:
                self.geo_col.hover_enter_obj(pymead_obj)
            else:
                self.setItemStyle(tree_item, "hovered")

        # Assign the current tree widget item to the previous item hovered
        self.previous_item_hovered = tree_item

    def leaveEvent(self, a0):
        """
        Reimplement the leave event to handle the case where the mouse exits directly sideways through the "Edit"
        button. In this case, the mouseMoveEvent will not catch the hover leave, so we need to put that logic here.
        """
        if self.previous_item_hovered is not None and self.previous_item_hovered.hoverable:
            pymead_obj = self.getPymeadObjFromItem(self.previous_item_hovered)
            if pymead_obj is not None:
                self.geo_col.hover_leave_obj(pymead_obj)
            else:
                self.setItemStyle(self.previous_item_hovered, "default")
            self.previous_item_hovered = None

    def addPymeadTreeItem(self, pymead_obj: PymeadObj):
        top_level_item = self.items[self.topLevelDict[pymead_obj.sub_container]]
        child_item = PymeadTreeWidgetItem([pymead_obj.name()])
        top_level_item.addChild(child_item)
        pymead_obj.tree_item = child_item

        if isinstance(pymead_obj, Param):
            right_column_widget = ValueSpin(self, pymead_obj)
            self.setItemWidget(child_item, 1, right_column_widget)

    def removePymeadTreeItem(self, pymead_obj: PymeadObj):
        top_level_item = self.items[self.topLevelDict[pymead_obj.sub_container]]
        top_level_item.removeChild(pymead_obj.tree_item)
        pymead_obj.tree_item = None

    def onExpandPressed(self):
        self.expandAll()

    def onCollapsePressed(self):
        self.collapseAll()

    def setItemStyle(self, item: PymeadTreeWidgetItem, style: str):
        valid_styles = ["default", "hovered"]
        if style not in ["default", "hovered"]:
            raise ValueError(f"Style found ({style}) is not a valid style. Must be one of {valid_styles}.")

        background_color = self.palette().color(self.backgroundRole())
        if style == "default":
            # item.setBackground(0, background_color)
            # item.setBackground(1, background_color)
            brush = QBrush(QColor(self.parent().parent().themes[self.parent().parent().current_theme]['main-color']))
            item.setForeground(0, brush)
        elif style == "hovered" and item.hoverable:
            # gradient = QtGui.QLinearGradient(0, 0, 150, 0)
            # gradient.setColorAt(0, QColor("#2678c9aa"))
            # gradient.setColorAt(1, self.palette().color(self.backgroundRole()))
            # item.setBackground(0, gradient)
            # item.setBackground(1, self.palette().color(self.backgroundRole()))
            brush = QBrush(QColor("#edb126"))
            item.setForeground(0, brush)

    def setForegroundColorAllItems(self, color: str):
        it = QTreeWidgetItemIterator(self)
        while it.value():
            it.value().setForeground(0, QBrush(QColor(color)))
            it += 1

    @staticmethod
    def undoRedoAction(action: typing.Callable):
        def wrapped(self, *args, **kwargs):
            self.gui_obj.undo_stack.append(deepcopy(self.geo_col.get_dict_rep()))
            action(self, *args, **kwargs)

        return wrapped

    @undoRedoAction
    def removeObjects(self, pymead_objs: typing.List[PymeadObj]):
        for pymead_obj in pymead_objs:
            self.geo_col.remove_pymead_obj(pymead_obj)

    @undoRedoAction
    def promoteParamsToDesvar(self, pymead_objs: typing.List[Param]):
        for pymead_obj in pymead_objs:
            self.geo_col.promote_param_to_desvar(pymead_obj)

    @undoRedoAction
    def demoteDesvarsToParam(self, pymead_objs: typing.List[DesVar]):
        for pymead_obj in pymead_objs:
            self.geo_col.demote_desvar_to_param(pymead_obj)

    @undoRedoAction
    def exposePointXY(self, pymead_objs: typing.List[Point]):
        for pymead_obj in pymead_objs:
            self.geo_col.expose_point_xy(pymead_obj)

    @undoRedoAction
    def coverPointXY(self, pymead_objs: typing.List[Param]):
        points_to_cover = list(set([pymead_obj.point for pymead_obj in pymead_objs]))
        for point_to_cover in points_to_cover:
            self.geo_col.cover_point_xy(point_to_cover)

    @undoRedoAction
    def equateConstraints(self, pymead_objs: typing.List[GeoCon]):
        for pymead_obj in pymead_objs[1:]:
            self.geo_col.equate_constraints(pymead_objs[0], pymead_obj)

    @undoRedoAction
    def splitBezier(self, pymead_objs: typing.List[Bezier]):
        for pymead_obj in pymead_objs:
            if not isinstance(pymead_obj, Bezier):
                continue
            pymead_obj.split(0.5)

    def contextMenuEvent(self, a0):
        # item = self.itemAt(a0.x(), a0.y())
        # if item is None:
        #     return

        items = self.selectedItems()
        if len(items) == 0:
            return

        if len(items) == 1 and items[0].text(0) == "Design Variables":
            menu = QMenu(self)
            addDesVarAction = menu.addAction("Add Design Variable")
            res = menu.exec_(a0.globalPos())

            if res is None:
                return

            if res is addDesVarAction:
                self.geo_col.add_desvar(0.0, "dv")

        elif len(items) == 1 and items[0].text(0) == "Parameters":
            menu = QMenu(self)
            addParameterAction = menu.addAction("Add Parameter")
            res = menu.exec_(a0.globalPos())

            if res is None:
                return

            if res is addParameterAction:
                self.geo_col.add_param(0.0, "param")

        elif all([item.parent() is not None for item in items]) and all(
                [item.parent() is items[0].parent() for item in items]):
            # button = self.itemWidget(item, 1)
            pymead_objs = [self.getPymeadObjFromItem(item) for item in items]
            pymead_objs = [pymead_obj for pymead_obj in pymead_objs if pymead_obj is not None]

            promoteAction = None
            demoteAction = None
            exposeAction = None
            coverAction = None
            addBezierPointAction = None
            splitBezierAction = None
            equateConstraintsAction = None

            pymead_obj_type = type(pymead_objs[0])

            menu = QMenu(self)
            if pymead_obj_type in [Param, LengthParam, AngleParam]:
                promoteAction = menu.addAction("Promote to Design Variable")
                if all([pymead_obj.point for pymead_obj in pymead_objs]):
                    coverAction = menu.addAction("Cover x and y Parameters")
            elif pymead_obj_type in [DesVar, LengthDesVar, AngleDesVar]:
                demoteAction = menu.addAction("Demote to Parameter")
                if all([pymead_obj.point for pymead_obj in pymead_objs]):
                    coverAction = menu.addAction("Cover x and y Parameters")
            elif pymead_obj_type is Point:
                exposeAction = menu.addAction("Expose x and y Parameters")
            elif pymead_obj_type is Bezier:
                addBezierPointAction = menu.addAction("Insert Control Point")
                splitBezierAction = menu.addAction("Split Curve")
            elif pymead_obj_type in [DistanceConstraint, RelAngle3Constraint]:
                equateConstraintsAction = menu.addAction("Equate Constraints")
            removeObjectAction = menu.addAction("Delete")

            res = menu.exec_(a0.globalPos())

            if res is None:
                return

            if res is removeObjectAction:
                self.removeObjects(pymead_objs)
            elif res is promoteAction:
                self.promoteParamsToDesvar(pymead_objs)
            elif res is demoteAction:
                self.demoteDesvarsToParam(pymead_objs)
            elif res is exposeAction:
                self.exposePointXY(pymead_objs)
            elif res is coverAction:
                self.coverPointXY(pymead_objs)
            elif res is addBezierPointAction:
                self.gui_obj.airfoil_canvas.addPointToCurve(pymead_objs[0].canvas_item)
            elif res is splitBezierAction:
                self.splitBezier(pymead_objs)
            elif res is equateConstraintsAction:
                self.equateConstraints(pymead_objs)

            self.geo_col.clear_selected_objects()
